import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.io.*;
import java.net.URL;

/**
   This is the main user interface for the non-html version of GameBoyEmu,
   it creates the main window which it shows and the DebugWindows which remain
   hidden until selected

   @author David Winchurch.
   @version 09/03/2001e
**/

public class GameBoyEmu extends Frame implements ActionListener, ItemListener, KeyListener, WindowListener
{
   //instance of GameBoyCPU. This is the virtual processor required to decypher the rom
   private GameBoyCPU CPU;

   //Parent Frame, Game Boy Graphic is attached to this
   //private Frame GBEmuFrame;

   //The 5 debug windows each responible for displaying internal info about the virtual gameboy
   private DebugWindow regDebug, //Displays the contents of the registers
                       sysDebug, //Displays internal system info
                       vidDebug, //Displays video subsystem info
                       romDebug, //Displays loaded rom info
                       palDebug; //Displays the paletes currently being used

   //A boolean array used to store the state of a key press
   private boolean[] KEYS_PRESSED={false,false,false,false,false,false,false,false};

   //An array used to set which key combinations do what allowing for configurable keys
   private int[] KEYS={KeyEvent.VK_UP, KeyEvent.VK_DOWN, KeyEvent.VK_LEFT, KeyEvent.VK_RIGHT, KeyEvent.VK_X, KeyEvent.VK_Z, KeyEvent.VK_ENTER, KeyEvent.VK_SHIFT, KeyEvent.VK_SPACE};

   //Integer used to store the video multiplier value. The display is scaled by this factor
   private int vidSize=1;

   //The image that will be drawn when the program is loaded
   private Image splashScreen, gameboy;

   //Menu System
   private MenuBar menu=new MenuBar();

   private Menu fileMenu=new Menu("File"),
                debugMenu=new Menu("Debug"),
                optionsMenu=new Menu("Options"),
                videoMenu=new Menu("Video");

   private MenuItem loadRom=new MenuItem("Load Rom"),
                    showAllDebug=new MenuItem("Show All Debug Info"),
                    hideAllDebug=new MenuItem("Hide All Debug Info"),
                    exit=new MenuItem("Exit"),
                    showRegDebug=new MenuItem("Show Registry Info"),
                    showSysDebug=new MenuItem("Show System Info"),
                    showVidDebug=new MenuItem("Show Video Info"),
                    showRomDebug=new MenuItem("Show Rom Info"),
                    showPalDebug=new MenuItem("Show Palette Info"),
                    vidSize1=new MenuItem("Size X 1"),
                    vidSize2=new MenuItem("Size X 2"),
                    paletteEditor=new MenuItem("Edit B&W Palette"),
                    keyConfig=new MenuItem("Configure Keys");

   private CheckboxMenuItem blurCheck=new CheckboxMenuItem("Blur Filter (Slow)"),
                            edgeEnhCheck=new CheckboxMenuItem("Edge Enhance (Slow)");

   //Unchangeable values for my reference so I know which position in the keys boolean
   //array to set to achieve the required changes
   private final int DIK_UP          = 0;
   private final int DIK_DOWN        = 1;
   private final int DIK_LEFT        = 2;
   private final int DIK_RIGHT       = 3;
   private final int DIK_A           = 4;
   private final int DIK_B           = 5;
   private final int DIK_START       = 6;
   private final int DIK_SELECT      = 7;
   private final int DIK_ALL_BUTTONS = 8;

   /*
      Constructor. Creates the Frame and adds all the required components to it.
      Creates and instantiates all the global values and objects
   */

   public GameBoyEmu()
   {
      //Calls the method which loads all the user defineable setting
      loadSettings();

      //Frame setup
      setTitle("GameBoyEmu");
      setLocation(100, 100);
      setResizable(false);
      setLayout(new BorderLayout());
      addWindowListener(this);
      addKeyListener(this);

      //Loads all the images used for the spinning cog animation on the file loading DebugWindow
      MediaTracker imageTracker = new MediaTracker(this);
      Image[] loadAnim=new Image[9];
      for(int i=0; i<9; i++)
      {
         //loads all the images into a MediaTracker
         loadAnim[i] = Toolkit.getDefaultToolkit().getImage("./images/cog"+Integer.toString(i+1)+".gif"); imageTracker.addImage(loadAnim[i] ,i);
      }
      //loads the splash screen image
      splashScreen = Toolkit.getDefaultToolkit().getImage("./images/screen.jpg");
      imageTracker.addImage(splashScreen ,9);
      //loads the gameboy image
      gameboy=Toolkit.getDefaultToolkit().getImage("./images/gameboy.gif");
      imageTracker.addImage(gameboy ,10);
      //Waits untill all the images have loaded
      try {imageTracker.waitForAll();}
      catch(InterruptedException e){}

      //Creates a new GameBoyCPU
      CPU=new GameBoyCPU(KEYS_PRESSED, loadAnim);

      //Gets the required handle on the GameBoyVideo instance created in GameBoyCPU
      CPU.getVideo().setVidMult(vidSize);
      CPU.getVideo().setBlur(blurCheck.getState());
      CPU.getVideo().setEdgeEnhance(edgeEnhCheck.getState());
      add(CPU.getVideo());
      //CPU.getVideo().addKeyListener(this);

      //Debug Window creation. They require access much of the variables held in the virtual CPU so this is passed to their constructors
      regDebug=new DebugWindow(this, "Registry Debug Window", CPU, DebugWindow.REG_INFO); regDebug.addWindowListener(this);
      sysDebug=new DebugWindow(this, "System Debug Window", CPU.getGBMem(), DebugWindow.SYS_INFO); sysDebug.addWindowListener(this);
      vidDebug=new DebugWindow(this, "Video Debug Window", CPU.getVideo(), DebugWindow.VID_INFO); vidDebug.addWindowListener(this);
      romDebug=new DebugWindow(this, "Rom Debug Info", CPU.getGBMem(), DebugWindow.ROM_INFO); romDebug.addWindowListener(this);
      palDebug=new DebugWindow(this, "Palette Debug Info", CPU.getVideo(), DebugWindow.PAL_INFO); palDebug.addWindowListener(this);

      //The palette config menuitem should only be selectable if a rom has been loadaed
      paletteEditor.setEnabled(false);

      //Menu Creation
      loadRom.addActionListener(this);
      showAllDebug.addActionListener(this);
      hideAllDebug.addActionListener(this);
      exit.addActionListener(this);
      showRegDebug.addActionListener(this);
      showSysDebug.addActionListener(this);
      showVidDebug.addActionListener(this);
      showRomDebug.addActionListener(this);
      showPalDebug.addActionListener(this);
      vidSize1.addActionListener(this);
      vidSize2.addActionListener(this);
      blurCheck.addItemListener(this);
      edgeEnhCheck.addItemListener(this);
      keyConfig.addActionListener(this);
      paletteEditor.addActionListener(this);

      //Sets up the Ctrl * shootcut keys for the MenuItems
      loadRom.setShortcut(new MenuShortcut(KeyEvent.VK_L));
      exit.setShortcut(new MenuShortcut(KeyEvent.VK_X));
      showAllDebug.setShortcut(new MenuShortcut(KeyEvent.VK_D));
      hideAllDebug.setShortcut(new MenuShortcut(KeyEvent.VK_H));
      vidSize1.setShortcut(new MenuShortcut(KeyEvent.VK_1));
      vidSize2.setShortcut(new MenuShortcut(KeyEvent.VK_2));
      blurCheck.setShortcut(new MenuShortcut(KeyEvent.VK_B));
      edgeEnhCheck.setShortcut(new MenuShortcut(KeyEvent.VK_E));
      keyConfig.setShortcut(new MenuShortcut(KeyEvent.VK_K));
      paletteEditor.setShortcut(new MenuShortcut(KeyEvent.VK_P));

      //Adds the menu objects in the right order
      fileMenu.add(loadRom);
      fileMenu.addSeparator();
      fileMenu.add(exit);
      debugMenu.add(showAllDebug);
      debugMenu.addSeparator();
      debugMenu.add(showRegDebug);
      debugMenu.add(showSysDebug);
      debugMenu.add(showVidDebug);
      debugMenu.add(showRomDebug);
      debugMenu.add(showPalDebug);
      debugMenu.addSeparator();
      debugMenu.add(hideAllDebug);
      videoMenu.add(vidSize1);
      videoMenu.add(vidSize2);
      videoMenu.addSeparator();
      videoMenu.add(blurCheck);
      videoMenu.add(edgeEnhCheck);
      optionsMenu.add(videoMenu);
      optionsMenu.add(keyConfig);
      optionsMenu.add(paletteEditor);
      menu.add(fileMenu);
      menu.add(optionsMenu);
      menu.add(debugMenu);

      //Frame set up
      setMenuBar(menu);
      pack();
      pack();
      setVisible(true);
   }

   /**
      This is the main procedure called when GameBoyEmu is run by Java.

      @Param args The command line arguments
   **/

   public static void main(String[] args)
   {
      GameBoyEmu instance=new GameBoyEmu();
   }

   /**
      Overwritten paint method. Ensures that the correct image is drawn
      when the program is obscured then revealed

      @param g the specified context to use for updating.
   **/

   public void paint(Graphics g)
   {
      //Draws the splash screen
      if(vidSize==1)
         CPU.getVideo().getGraphics().drawImage(splashScreen,0,0,320,288,this);
      //Drawing to the screen when the CPU is running can cause crashes
      if(!CPU.getThreadState() && !getTitle().equals("GameBoyEmu"))
         CPU.getVideo().refresh();
   }

   /**
      Runs all the methods required on closure of the program including
      saving the setting and and save data the rom created
   **/

   public void exit()
   {
      CPU.stopThread();
      saveSettings();
      CPU.saveCartRam();
      CPU.savePalleteInfo();
      dispose();
      System.gc();
      System.exit(0);
   }

   /**
      Loads user configurable settings from a text file
   **/

   public void loadSettings()
   {
      try
      {
         //Creates a new BufferedReader from a file on the hard disk. Throws an error if the file is not present
         BufferedReader reader=new BufferedReader(new FileReader("./GameBoyEmu.ini"));

         //Attemps to read in settings from the loaded file throws an error if it fails
         try
         {
            //Used to track the number of lines in the text file
            int fileSize=0;
            //Loops until the end of the file
            while (reader.readLine()!=null) fileSize++;

            //Recreates the BufferedReader. Used to start the reader at the begining of the file again
            reader=new BufferedReader(new FileReader("./GameBoyEmu.ini"));

            //loops through the file reading in each line and identifying it as a user input line
            for(int i=0; i<fileSize; i++)
            {
               String temp=reader.readLine();
               //loops through the string checking each character at a time to see if it's an '='
               for(int j=0; j<temp.length(); j++)
               {
                  //once it finds the required character these if statements find out if the subtring before
                  //the '=' is part of the stored user information
                  if(temp.charAt(j)=='=')
                  {
                     if(temp.substring(0, j).equals("up")) KEYS[DIK_UP]=Integer.parseInt(temp.substring(j+1, temp.length()));
                     if(temp.substring(0, j).equals("down")) KEYS[DIK_DOWN]=Integer.parseInt(temp.substring(j+1, temp.length()));
                     if(temp.substring(0, j).equals("left")) KEYS[DIK_LEFT]=Integer.parseInt(temp.substring(j+1, temp.length()));
                     if(temp.substring(0, j).equals("right")) KEYS[DIK_RIGHT]=Integer.parseInt(temp.substring(j+1, temp.length()));

                     if(temp.substring(0, j).equals("a")) KEYS[DIK_A]=Integer.parseInt(temp.substring(j+1, temp.length()));
                     if(temp.substring(0, j).equals("b")) KEYS[DIK_B]=Integer.parseInt(temp.substring(j+1, temp.length()));
                     if(temp.substring(0, j).equals("start")) KEYS[DIK_START]=Integer.parseInt(temp.substring(j+1, temp.length()));
                     if(temp.substring(0, j).equals("select")) KEYS[DIK_SELECT]=Integer.parseInt(temp.substring(j+1, temp.length()));

                     if(temp.substring(0, j).equals("all")) KEYS[DIK_ALL_BUTTONS]=Integer.parseInt(temp.substring(j+1, temp.length()));

                     if(temp.substring(0, j).equals("vidsize")) vidSize=Integer.parseInt(temp.substring(j+1, temp.length()));
                     if(temp.substring(0, j).equals("blur")) {if(temp.substring(j+1, temp.length()).equals("true")) blurCheck.setState(true);}
                     if(temp.substring(0, j).equals("edgeenh")) {if(temp.substring(j+1, temp.length()).equals("true")) edgeEnhCheck.setState(true);}
                  }
               }
            }
            //Closes the BufferedReader
            reader.close();
         }
         catch(IOException ex){System.out.println("Error reading file");}
      }
      catch(FileNotFoundException ex){System.out.println("no file");}
   }

   /**
      Writes user configurable information into a text file
   **/

   public void saveSettings()
   {
      try
      {
         //Attempts to create a new file on the hard disk to write strings into
         //throws an IOException should it fail
         BufferedWriter writer=new BufferedWriter(new FileWriter("./GameBoyEmu.ini"));

         //Writes all the nessesary information into the text file
         writer.write("up=" + KEYS[DIK_UP]); writer.newLine();
         writer.write("down=" + KEYS[DIK_DOWN]); writer.newLine();
         writer.write("left=" + KEYS[DIK_LEFT]); writer.newLine();
         writer.write("right=" + KEYS[DIK_RIGHT]); writer.newLine();

         writer.write("a=" + KEYS[DIK_A]); writer.newLine();
         writer.write("b=" + KEYS[DIK_B]); writer.newLine();
         writer.write("start=" + KEYS[DIK_START]); writer.newLine();
         writer.write("select=" + KEYS[DIK_SELECT]); writer.newLine();

         writer.write("all=" + KEYS[DIK_ALL_BUTTONS]); writer.newLine();
         writer.write("vidsize=" + vidSize); writer.newLine();
         writer.write("blur=" + blurCheck.getState()); writer.newLine();
         writer.write("edgeenh=" + edgeEnhCheck.getState());

         //Closes the file
         writer.close();
      }
      catch(IOException ex){System.out.println("no file");}
   }

   /**
      Gets the visible state of the DebugWindows.

      @returns boolean true if any of the DebugWindows are visible
   **/

   public boolean getDebug()
   {
      return (regDebug.isVisible() || sysDebug.isVisible() || vidDebug.isVisible() || romDebug.isVisible() || palDebug.isVisible())? true:false;
   }

   /**
      calls the refresh() method in all the DebugWindows except the rom Info Window
      as that doesnt change once the rom is loaded
   **/

   public void debugRefresh()
   {
      if(regDebug.isVisible()) regDebug.refresh();
      if(sysDebug.isVisible()) sysDebug.refresh();
      if(vidDebug.isVisible()) vidDebug.refresh();
      if(palDebug.isVisible()) palDebug.refresh();
   }

   /**
      This is the ActionEvent handler for GameBoyEmu. It accepts events generated
      by any component with an ActionListener added.

      @param event The event that was generated.
   **/

   public void actionPerformed(ActionEvent event)
   {
      //gets the object that generated the ActionEvent for identification purposes
      Object source=event.getSource();

      //The if statements Discover which component generated the ActionEvent and runs the appropriate methods

      //Run when the loadrom MenuItem is selected
      if(source==loadRom)
      {
         //Prepares the CPU to accept a new rom by stopping it and saving any rom data and palette info
         CPU.stopThread();
         CPU.saveCartRam();
         CPU.savePalleteInfo();

         //Generates a default file loading window
         FileDialog loadFile=new FileDialog(this,"Select Game Boy Rom",FileDialog.LOAD);
         //Sets the initial directory
         loadFile.setDirectory("D:\\Programs\\Java\\GameBoyEmu\\Roms\\");
         //Sets which file types to display
         loadFile.setFile("*.cgb;*.CGB;*.gbc;*.GBC;*.gb;*.GB");
         loadFile.show();

         if(vidSize==1)
            CPU.getVideo().getGraphics().drawImage(splashScreen,0,0,320,288,this);

         //If the loading of a rom was succesful update the DebugWindows and reset the virtual CPU
         if(CPU.loadRom(loadFile.getDirectory()+loadFile.getFile()))
         {
            setTitle(loadFile.getFile());

            //The palete editor menuitem should only be selectable of the rom is not a color rom
            if(CPU.getMem(0x143)>=0x80)
               paletteEditor.setEnabled(false);
            else
               paletteEditor.setEnabled(true);
               
            CPU.fullReset();
            romDebug.refresh();
            debugRefresh();
         }

         //If there are no visible DebugWindows and a rom has actually been loaded start the virtual CPU
         if(!getDebug() && !getTitle().equals("GameBoyEmu")) CPU.startThread();
      }
      //Updates the variables and frame to accept the new display size
      else if(source==vidSize1 && vidSize==1) {repaint();}
      else if(source==vidSize1 && vidSize!=1) {vidSize=1; CPU.getVideo().setVidMult(1); repaint(50);}
      else if(source==vidSize2 && vidSize!=2) {vidSize=2; CPU.getVideo().setVidMult(2);}
      //Opens the key Configuration window
      else if(source==keyConfig) {KeyConfigurator tempKeyConfig = new KeyConfigurator(this, KEYS, gameboy);}
      //Opens the palette editor window
      else if(source==paletteEditor) {PaletteEditor tempPalEdit = new PaletteEditor(this, CPU);}
      else if(source==exit) exit();
      //The followind deals with the DebugWindows stoping the virtual CPU when they are visible and starting
      //it again when they are closed
      else if(source==showAllDebug) {CPU.stopThread(); regDebug.setVisible(true); sysDebug.setVisible(true); vidDebug.setVisible(true); romDebug.setVisible(true); palDebug.setVisible(true); debugRefresh(); repaint();}
      else if(source==showRegDebug) {CPU.stopThread(); regDebug.setVisible(true); regDebug.refresh(); repaint();}
      else if(source==showSysDebug) {CPU.stopThread(); sysDebug.setVisible(true); sysDebug.refresh(); repaint();}
      else if(source==showVidDebug) {CPU.stopThread(); vidDebug.setVisible(true); vidDebug.refresh(); repaint();}
      else if(source==showRomDebug) {CPU.stopThread(); romDebug.setVisible(true); repaint();}
      else if(source==showPalDebug) {CPU.stopThread(); palDebug.setVisible(true); palDebug.refresh(); repaint();}
      else if(source==hideAllDebug) {regDebug.setVisible(false); sysDebug.setVisible(false); vidDebug.setVisible(false); romDebug.setVisible(false); palDebug.setVisible(false); if(!getTitle().equals("GameBoyEmu")) CPU.startThread();}
   }

   /**
      This is the ItemEvent handler for GameBoyEmu. It accepts events generated
      by any component with an ItemListener added.

      @param event The event that was generated.
   **/

   public void itemStateChanged(ItemEvent event)
   {
      Object source=event.getSource();

      if(source==blurCheck)
      {
         //Ensures that only one filter can be applied at once
         if(blurCheck.getState())
         {
            CPU.getVideo().setEdgeEnhance(false);
            edgeEnhCheck.setState(false);
         }
         CPU.getVideo().setBlur(blurCheck.getState());
         //redraws the screen if the thread has been stopped
         if(!CPU.getThreadState())
            CPU.getVideo().refresh();
      }
      if(source==edgeEnhCheck)
      {
         //Ensures that only one filter can be applied at once
         if(edgeEnhCheck.getState())
         {
            CPU.getVideo().setBlur(false);
            blurCheck.setState(false);
         }
         CPU.getVideo().setEdgeEnhance(edgeEnhCheck.getState());
         //redraws the screen if the thread has been stopped
         if(!CPU.getThreadState())
            CPU.getVideo().refresh();
      }
   }

   /**
      This is the key pressed event handler for GameBoyEmu. It accepts KeyEvents generated
      by pressing a key when any component with a KeyListener has focus

      @param k The KeyEvent that was generated.
   **/

   public void keyPressed(KeyEvent k)
   {
      //gets the integer corresponding to the pressed key
      int keypress=k.getKeyCode();

      //checks to see if the correct keys have been pressed and sets the corresponding boolean to true
      if (KEYS[DIK_UP]==keypress) {KEYS_PRESSED[DIK_UP]=true;}
      else if (KEYS[DIK_DOWN]==keypress) {KEYS_PRESSED[DIK_DOWN]=true;}
      else if (KEYS[DIK_LEFT]==keypress) {KEYS_PRESSED[DIK_LEFT]=true;}
      else if (KEYS[DIK_RIGHT]==keypress) {KEYS_PRESSED[DIK_RIGHT]=true;}

      else if (KEYS[DIK_A]==keypress) {KEYS_PRESSED[DIK_A]=true;}
      else if (KEYS[DIK_B]==keypress) {KEYS_PRESSED[DIK_B]=true;}
      else if (KEYS[DIK_START]==keypress) {KEYS_PRESSED[DIK_START]=true;}
      else if (KEYS[DIK_SELECT]==keypress) {KEYS_PRESSED[DIK_SELECT]=true;}

      else if (KEYS[DIK_ALL_BUTTONS]==keypress) {KEYS_PRESSED[DIK_A]=true; KEYS_PRESSED[DIK_B]=true; KEYS_PRESSED[DIK_START]=true; KEYS_PRESSED[DIK_SELECT]=true;}

      //If enter is pressed and the Debug Windows are showing perform one opCode
      if(!CPU.getThreadState() && !getTitle().equals("GameBoyEmu") && getDebug() && keypress==KeyEvent.VK_ENTER)
      {
         CPU.opCodeList1();
         debugRefresh();
      }
   }

   /**
      This is the key released event handler for GameBoyEmu. It accepts KeyEvents generated
      by releasing a key when any component with a KeyListener has focus

      @param k The KeyEvent that was generated.
   **/

   public void keyReleased(KeyEvent k)
   {
      //gets the integer corresponding to the released key
      int keypress=k.getKeyCode();

      //checks to see if the correct keys have been released and sets the corresponding boolean to false
      if (KEYS[DIK_UP]==keypress) KEYS_PRESSED[DIK_UP]=false;
      else if (KEYS[DIK_DOWN]==keypress) KEYS_PRESSED[DIK_DOWN]=false;
      else if (KEYS[DIK_LEFT]==keypress) KEYS_PRESSED[DIK_LEFT]=false;
      else if (KEYS[DIK_RIGHT]==keypress) KEYS_PRESSED[DIK_RIGHT]=false;

      else if (KEYS[DIK_A]==keypress) KEYS_PRESSED[DIK_A]=false;
      else if (KEYS[DIK_B]==keypress) KEYS_PRESSED[DIK_B]=false;
      else if (KEYS[DIK_START]==keypress) KEYS_PRESSED[DIK_START]=false;
      else if (KEYS[DIK_SELECT]==keypress) KEYS_PRESSED[DIK_SELECT]=false;

      else if (KEYS[DIK_ALL_BUTTONS]==keypress) {KEYS_PRESSED[DIK_A]=false; KEYS_PRESSED[DIK_B]=false; KEYS_PRESSED[DIK_START]=false; KEYS_PRESSED[DIK_SELECT]=false;}

   }

   /*
      This is the key typed event handler for GameBoyEmu. It is here because java requires you
      to implement it for KeyListeners to work. Dos nothing in this program

      @param k The KeyEvent that was generated.
   */

   public void keyTyped(KeyEvent k) {}

   /**
      This is the window event handler for GameBoyEmu. It accepts WindowEvents generated when
      a frame with an attached WindowListener is closed

      @param e The WindowEvent that was generated.
   **/

   public void windowClosing(WindowEvent e)
   {
      Object source=e.getSource();

      if(source==this) exit();
      else
      {
         //Closes the specified debugWindow and restarts the virtual processor if necessary
         ((DebugWindow)(source)).dispose();
         if(!getDebug() && !getTitle().equals("GameBoyEmu")) CPU.startThread();
      }
   }

   /*
      The following methods are here because they are required when a WindowListener is
      implemented. They do nothing in this program

      @param e The WindowEvent that was generated.
   */

   public void windowOpened(WindowEvent e) {}
   public void windowClosed(WindowEvent e) {}
   public void windowIconified(WindowEvent e) {}
   public void windowDeiconified(WindowEvent e) {}
   public void windowActivated(WindowEvent e) {}
   public void windowDeactivated(WindowEvent e) {}  
} 
