import java.awt.*;
import java.awt.image.*;
import java.awt.image.BufferedImage;

/**
   This class is responsible for decoding the way the graphics work inside the gameboy
   and converting it so that it can be displayed on java compatable computers

   @author David Winchurch
   @version 11/03/2001e
**/

public class GameBoyVideo extends Component
{
   //pointers to relevant areas in the virtual gameboys memory arrays
   private char GB_MEM[], VRAM[];

   //The image that will be drawn to the screen
   private Image screenImage;

   //The graphics surface to draw too
   private Graphics screenGraphics;

   //Instance of a fast image producing class
   private DirectImage dImage;

   //The filter matricies
   private float[] BLUR3x3 = {0.05f, 0.05f, 0.05f, 0.05f, 0.6f, 0.05f, 0.05f, 0.05f, 0.05f},
                   EDGEENHANCE3x3 = {-0.15f, -0.15f, -0.15f, -0.15f, 2.2f, -0.15f, -0.15f, -0.15f, -0.15f};

   //The filter kernels
   private Kernel blurKernel = new Kernel(3, 3, BLUR3x3),
                  edgeEnhKernel = new Kernel(3, 3, EDGEENHANCE3x3);

   //The conversion operations for the filters
   private ConvolveOp blurOp = new ConvolveOp(blurKernel, ConvolveOp.EDGE_NO_OP, null),
                      edgeEnhOp = new ConvolveOp(edgeEnhKernel, ConvolveOp.EDGE_NO_OP, null);

   //two buffered images used for filtering the displayed image
   private BufferedImage buffer = new BufferedImage(160, 144, BufferedImage.TYPE_INT_RGB),
                         buffer2 = new BufferedImage(160, 144, BufferedImage.TYPE_INT_RGB);

   //surface to copy the screen image to for filtering
   private Graphics2D buf = buffer.createGraphics();

   //The global variables required to keep track of the many things happening in the graphics subsystems
   private int xOffSet, yOffSet, imageWidth, imageHeight, winPoint, tileArea, bgPoint, vidMult, currentLine, scrollX, scrollY, tile, xTile, yTile, bitX, bitY, colour, charBnkSlct, tileProp, winX, winY, currentSpr, sprX, sprY, sprF, sprSize, sprPerLine, tx, ty;


   private int yLookUp[]    = new int[144],
               //array that stores which palette entry has been used for the corresponding pixel
               pixelColor[] = new int[160*144],
               //arrays that store the colours for each palette
               backPal[][]  = new int[4][8],
               forePal[][]  = new int[4][8];

   //array that stores the pixels priority, true says it is drawn above the sprite layer
   private boolean pixelPriority[]=new boolean[160*144],
                   blur, //toggles blur modes on/off
                   edgeEnh; //toggles edge enhance on/off

   //Unchangeable values that correspond to areas in the virtual memory
   private final int LCD_CTRL         = 0xFF40;
   private final int LCD_SCROLL_Y     = 0xFF42;
   private final int LCD_SCROLL_X     = 0xFF43;
   private final int LCD_Y_LOC        = 0xFF44;
   private final int LCD_WIN_Y        = 0xFF4A;
   private final int LCD_WIN_X        = 0xFF4B;

   /**
      Constructor. Creates and instantiates all the global variables

      @param gbMem pointer to the virtual memory array
      @param vram pointer to the virtual graphics memory
   **/

   public GameBoyVideo(char[] gbMem, char[] vram)
   {
      //provides global pointers to the virtual memory locations
      GB_MEM=gbMem;
      VRAM=vram;

      //Creates the image objects
      dImage=new DirectImage(160, 144);
      screenImage=createImage(dImage);

      //creates a lookup table to give the illusion of 2 dimentions in a 1d array
      for(int i=0; i<144; i++) yLookUp[i]=i*160;
   }

   /**
      ensures that any system calls to update do nothing. for optimization purposes

      @param g the specified context to use for updating.
   **/

   public void update(Graphics g) {}

   /**
      ensures that any system calls to repaint do nothing. for optimization purposes

      @param     tm   maximum time in milliseconds before update.
      @param     x    the <i>x</i> coordinate.
      @param     y    the <i>y</i> coordinate.
      @param     width    the width.
      @param     height   the height.
   **/

   public void repaint(long tm, int x, int y, int width, int height) {}

   /**
      ensures that any system calls to paintAll do nothing. for optimization purposes

      @param g the graphics context to use for painting.
   **/

   public void paintAll(Graphics g) {}

   /**
      Draws the image to the screen. applies a blur filter should one be required
   **/

   public void refresh()
   {
      try
      {
         //Only creates the graphics surface if its null. for optimization purposes
         if(screenGraphics==null) {screenGraphics=getGraphics();}

		     //Updates screenImage to include any changes
		     screenImage.flush();

		     if(blur)
		     {
			      //copies the screenImage to the Graphics2D surface for filtering
			      buf.drawImage(screenImage, 0, 0, null);

			      //Applies the filter
			      blurOp.filter(buffer, buffer2);

			      //draws the resulting blured image to the screen
			      screenGraphics.drawImage(buffer2, xOffSet, yOffSet, imageWidth, imageHeight, Color.white, this);
		     }
		     else if(edgeEnh)
		     {
			      //copies the screenImage to the Graphics2D surface for filtering
			      buf.drawImage(screenImage, 0, 0, null);

			      //Applies the filter
			      edgeEnhOp.filter(buffer, buffer2);

            //draws the resulting blured image to the screen
			      screenGraphics.drawImage(buffer2, xOffSet, yOffSet, imageWidth, imageHeight, Color.white, this);
         }
		     else
			      //draws the unaltered gameboy graphic to the screen
			      screenGraphics.drawImage(screenImage, xOffSet, yOffSet, imageWidth, imageHeight, Color.white, this);
      }
      catch(NullPointerException ex) {}
   }

   /**
      This method decodes the 64 color tile based graphics system used in the gameboy
      and converts it to a form displayable on the java virtual machine. It calculates each pixel and sets it
      in the image. Its very complex and I will not explain it here instead please refer to
      section FILL IN HERE!!!!!!!!!!!!!! for my explanation of the gameboy graphics system

      @returns the number of sprites on each line as an integer
   **/

   public int drawLine()
   {
      //checks the system is on a drawable line (143 lines out of 153 are visible)
      if(GB_MEM[LCD_Y_LOC]<0x90)
      {
         currentLine=GB_MEM[LCD_Y_LOC];
         sprPerLine=0;

         if((GB_MEM[LCD_CTRL]&0x01)>0)
         {
            scrollX=GB_MEM[LCD_SCROLL_X];
            scrollY=GB_MEM[LCD_SCROLL_Y];

            for(int x=0; x<160; x++)
            {
               //Sorts out the wrapping round of tiles
               if((scrollY+currentLine)>255) scrollY-=255;
               if((scrollX+x)>255) scrollX-=255;

               //works out which tile the pixel belongs to from the 32x32 tile map
               xTile=(scrollX+x)>>3;
               yTile=(scrollY+currentLine)>>3;

               //works out the x,y coordinate of the tile
               bitX=(scrollX+x)&0x07;
               bitY=(scrollY+currentLine)&0x07;

               //FOR COLOR SUPPORT gets the attributes of the current tile from the second vram bank
               tileProp=VRAM[0x2000+bgPoint+(yTile<<5)+xTile];

               //gets the number of the tile the pixel belongs to
               tile=VRAM[bgPoint+(yTile<<5)+xTile];

               //the tiles can be stored at two locations in memory. if it is 0x8800 the tile numbers are signed
               if(tileArea==0x800) tile^=0x80;

               //COLOR ONLY gameboy color tiles can be x,y flipped depending on bits set. This code caters for this
               if((tileProp&0x20)>0) bitX=(7-bitX);
               if((tileProp&0x40)>0) bitY=(7-bitY);

               //FOR COLOR SUPPORT gameboy color tile data can come from both vram banks depending on bits set
               charBnkSlct=(tileProp&0x08)>0? 0x2000:0;

               //Calculates which of the 4 positions in one of the palette arrays the pixel corresponds to
               colour=(((VRAM[charBnkSlct+tileArea+1+(bitY<<1)+(tile<<4)]>>(7-bitX))&0x01)<<1) | ((VRAM[charBnkSlct+tileArea+(bitY<<1)+(tile<<4)]>>(7-bitX))&0x01);

               //sets the colour of the pixel in the DirectImage array
               dImage.pixels[yLookUp[currentLine]+x]=backPal[colour][tileProp&0x07];

               //the following lines are used later to determine the pixel priority of the three layers
               pixelColor[yLookUp[currentLine]+x]=colour;
               pixelPriority[yLookUp[currentLine]+x]=(tileProp&0x80)>0? true:false;
            }
         }

         if((GB_MEM[LCD_CTRL]&0x20)>0 && (GB_MEM[0xFF4A]<=currentLine))
         {
            //the windows x position is offset by 7
            winX=GB_MEM[LCD_WIN_X]-7;
            winY=GB_MEM[LCD_WIN_Y];

            for(int wx=0; wx<(160-winX); wx++)
            {
               //works out which tile the pixel belongs to from the 32x32 tile map
               xTile=(wx)>>3;
               yTile=((currentLine-winY))>>3;

               //works out the x,y coordinate of the tile
               bitX=(wx)&0x07;
               bitY=((currentLine-winY))&0x07;

               //FOR COLOR SUPPORT gets the attributes of the current tile from the second vram bank
               tileProp=VRAM[0x2000+winPoint+(yTile<<5)+xTile];

               //gets the number of the tile the pixel belongs to
               tile=VRAM[winPoint+(yTile<<5)+xTile];

               //the tiles can be stored at two locations in memory. if it is 0x8800 the tile numbers are signed
               if(tileArea==0x800) tile^=0x80;

               //FOR COLOR SUPPORT gameboy color tile data can come from both vram banks depending on bits set
               charBnkSlct=(tileProp&0x08)>0? 0x2000:0;

               //Calculates which of the 4 positions in one of the palette arrays the pixel corresponds to
               colour=(((VRAM[charBnkSlct+tileArea+1+(bitY<<1)+(tile<<4)]>>(7-bitX))&0x01)<<1) | ((VRAM[charBnkSlct+tileArea+(bitY<<1)+(tile<<4)]>>(7-bitX))&0x01);

               if((wx+winX)<160 && (wx+winX)>=0)
               {
                  //sets the colour of the pixel in the DirectImage array provided its in the viewable area
                  dImage.pixels[yLookUp[currentLine]+(wx+winX)]=backPal[colour][tileProp&0x07];

                  //the following lines are used later to determine the pixel priority of the three layers
                  pixelColor[yLookUp[currentLine]+(wx+winX)]=colour;
                  pixelPriority[yLookUp[currentLine]+(wx+winX)]=(tileProp&0x80)>0? true:false;
               }
            }
         }

         if((GB_MEM[LCD_CTRL]&0x02)>0)
         {
            //finds out the size of the sprites used 8*8 or 8*16
            sprSize=((GB_MEM[LCD_CTRL]&0x04)>0? 15:7);

            for(int i=39; i>=0; i--)
            {
               currentSpr=i<<2;

               //gets the y position of the sprite (sprites are offset by 16 on the y axis)
               sprY=GB_MEM[(0xFE00+currentSpr)]-16;

               //checks to see if any part of the sprite is on the current line
               if(sprY<=currentLine && (sprY+sprSize)>=currentLine)
               {
                  //increments sprPerLine for each sprite it finds on this line. this valse
                  //is used for timing purposes in the virtual CPU
                  sprPerLine++;

                  //gets the y coordinate of the line of the sprite were working on
                  sprY=currentLine-sprY;

                  //gets the x position of the sprite (sprites are offset by 8 on the x axis)
                  sprX=GB_MEM[(0xFE01+currentSpr)]-8;

                  //gets the tile used by the sprite
                  tile=GB_MEM[(0xFE02+currentSpr)];

                  //gets the byte that contains the flags for the sprite
                  sprF=GB_MEM[(0xFE03+currentSpr)];

                  //flips the sprite on the y-axis if specified by the flags
                  if((sprF&0x40)>0) sprY=sprSize-sprY;

                  for(int x=0; x<8; x++)
                  {
                     //gets the actual x-coordinate screen position of the sprite pixel taking into account x-axis flipping
                     tx=((sprF&0x20)>0? sprX+(7-x):sprX+x);

                     //logic for deciding if the sprite pixel should be displayed or not
                     if(tx>=0 && tx<160 && (!pixelPriority[yLookUp[currentLine]+tx] || (pixelPriority[yLookUp[currentLine]+tx] && pixelColor[yLookUp[currentLine]+tx]==0)) && ((sprF&0x80)==0 || ((sprF&0x80)>0 && pixelColor[yLookUp[currentLine]+tx]==0)))
                     {
                        //FOR COLOR SUPPORT gameboy color tile data can come from both vram banks depending on bits set
                        charBnkSlct=(sprF&0x08)>0? 0x2000:0;

                        //Calculates which of the 4 positions in one of the palette arrays the pixel corresponds to
                        colour=(((VRAM[charBnkSlct+1+(sprY<<1)+(tile<<4)]>>(7-x))&0x01)<<1) | ((VRAM[charBnkSlct+(sprY<<1)+(tile<<4)]>>(7-x))&0x01);

                        //colour 0 is always transparent on sprites so that they dont appear as a character in a box
                        if(colour!=0)
                           dImage.pixels[yLookUp[currentLine]+tx]=forePal[colour][GB_MEM[0x143]>=0x80? sprF&0x07:((sprF&0x10)>0? 1:0)];
                     }
                  }
               }
            }
         }
      }
      return sprPerLine>10? 10:sprPerLine;
   }

   /**
      Gets the preferred size of this component.
      @return A dimension object indicating this component's preferred size.
   **/

   public Dimension getPreferredSize() {return new Dimension(320, 288);}

   /**
      Gets the specified background colour from the background palette array

      @return A color from the background palette
   **/

   public Color getBackColour(int loc, int pal) {return new Color(backPal[loc][pal]);}

   /**
      Gets the specified foreground colour from the foreground palette array

      @return A color from the foreground palette
   **/

   public Color getForeColour(int loc, int pal) {return new Color(forePal[loc][pal]);}

   /**
      Sets the multiplication factor applied to the screenImage

      @param mult the desired multiplication factor
   **/

   public void setVidMult(int mult)
   {
      vidMult=mult;
      xOffSet=160-((160*vidMult)/2);
      yOffSet=144-((144*vidMult)/2);
      imageWidth=160*vidMult;
      imageHeight=144*vidMult;
   }

   /**
      Sets the state of the blur filter

      @param b true if the blur filter should be applied
   **/

   public void setBlur(boolean b) {blur=b;}

   /**
      Sets the state of the Edge Enhance filter

      @param b true if the Edge Enhance filter should be applied
   **/

   public void setEdgeEnhance(boolean b) {edgeEnh=b;}

   /**
      Adds a color to the specified location in the background palette array

      @param loc the color number between 0-3
      @param pal the palette number between 0-7
      @paran c the specified Color
   **/

   public void setBackColour(int loc, int pal, Color c)
   {
      if(loc>backPal[0].length || pal>backPal[1].length)
         System.out.println("You have tried to set a background palette color outside of normal boundaries");
      else
         backPal[loc][pal]=c.getRGB();
   }

   /**
      Adds a color to the specified location in the foreground palette array

      @param loc the color number between 0-3
      @param pal the palette number between 0-7
      @paran c the specified Color
   **/

   public void setForeColour(int loc, int pal, Color c)
   {
      if(loc>backPal[0].length || pal>backPal[1].length)
         System.out.println("You have tried to set a foreground palette color outside of normal boundaries");
      else
         forePal[loc][pal]=c.getRGB();
   }

   /**
      Sets the position in Vram from which the background tile pointers come from

      @param p the start of the background tile pointers data area
   **/

   public void setBGPointers(int p) {bgPoint=p;}

   /**
      Sets the position in Vram from which the window tile pointers come from

      @param p the start of the window tile pointers data area
   **/

   public void setWinPointers(int p) {winPoint=p;}

   /**
      Sets the position in Vram from which the tile data comes from

      @param p the start of the tile data area
   **/

   public void setTileArea(int p) {tileArea=p;}

   /**
      Gets the on/off state of the LCD screen

      @return String specifying the state of the LCD screen
   **/

   public String getLCD() {if((GB_MEM[LCD_CTRL]&0x80)>0) return "On"; else return "Off";}

   /**
      Gets the on/off state of the window layer

      @return String specifying the state of the window layer
   **/

   public String getWIN() {if((GB_MEM[LCD_CTRL]&0x20)>0) return "On"; else return "Off";}

   /**
      Gets the on/off state of the background layer

      @return String specifying the state of the background layer
   **/

   public String getBG() {if((GB_MEM[LCD_CTRL]&0x01)>0) return "On"; else return "Off";}

   /**
      Gets the on/off state of the objects layer

      @return String specifying the state of the objects layer
   **/

   public String getOBJ() {if((GB_MEM[LCD_CTRL]&0x02)>0) return "On"; else return "Off";}

   /**
      Gets the location of the tile data in vram

      @return String specifying the location of the tile data area in vram in hex
   **/

   public String getTileArea() {return new String(Integer.toHexString(tileArea+0x8000) + " - " +  Integer.toHexString(tileArea+0x8FFF));}

   /**
      Gets the location of the background tile pointers in vram

      @return String specifying the location of the background tile pointes in vram in hex
   **/

   public String getBGPointers() {return new String(Integer.toHexString(bgPoint+0x8000) + " - " +  Integer.toHexString(bgPoint+0x83FF));}

   /**
      Gets the location of the window tile pointers in vram

      @return String specifying the location of the window tile pointes in vram in hex
   **/

   public String getWinPointers() {return new String(Integer.toHexString(winPoint+0x8000) + " - " + Integer.toHexString(winPoint+0x83FF));}

   /**
      Gets the size of the sprites

      @return The size of the sprites as a String
   **/

   public String getOBJSIZE() {if ((GB_MEM[LCD_CTRL]&0x04)>0) return "8*16"; else return "8*8";}

   /**
      Gets the Y scroll position of the background

      @return The Y scroll position of the background as a hexadecimal String
   **/

   public String getScrollY() {return Integer.toString((int)(GB_MEM[LCD_SCROLL_Y]));}

   /**
      Gets the X scroll position of the background

      @return The X scroll position of the background as a hexadecimal String
   **/

   public String getScrollX() {return Integer.toString((int)(GB_MEM[LCD_SCROLL_X]));}

   /**
      Gets the current horizontal scanline

      @return The Y scroll position of the background as a hexadecimal String
   **/

   public String getCurrentLine() {return Integer.toString((int)(GB_MEM[LCD_Y_LOC]));}

   /**
      Gets the Y scroll position of the window

      @return The Y scroll position of the window as a hexadecimal String
   **/

   public String getWinY() {return Integer.toString((int)(GB_MEM[LCD_WIN_Y]));}

   /**
      Gets the X scroll position of the window

      @return The X scroll position of the window as a hexadecimal String
   **/

   public String getWinX() {return Integer.toString((int)(GB_MEM[LCD_WIN_X]));}
}